@@ -75,6 +75,7 @@ struct ast_pointer_type {
 
 struct ast_tagged_union_type {
 	struct ast_type *type;
+	bool unwrap;
 	struct ast_tagged_union_type *next;
 };
 
@@ -85,8 +86,7 @@ struct ast_tuple_type {
 
 struct ast_struct_union_field {
 	struct ast_struct_union_field *next;
-	struct ast_expression *offset;
-	const char *name;
+	const char *name; // null if embed, may be "_"
 	struct ast_type *type;
 };
 
@@ -98,9 +98,9 @@ struct ast_struct_union_type {
 struct ast_type {
 	struct location loc;
 	enum type_storage storage;
-	unsigned int flags;
 	union {
 		struct ast_array_type array;
+		struct ast_type *error;
 		struct ast_function_type func;
 		struct ast_pointer_type pointer;
 		struct ast_slice_type slice;
@@ -109,10 +109,8 @@ struct ast_type {
 		struct ast_tuple_type tuple;
 		struct {
 			struct ident *alias;
-			union {
-				struct ast_enum_type _enum;
-				bool unwrap;
-			};
+			// Only valid for enums
+			struct ast_enum_type _enum;
 		};
 	};
 };
@@ -167,6 +165,7 @@ struct ast_expression_assert {
 
 struct ast_expression_assign {
 	enum binarithm_operator op;
+	// object == NULL for discarding assignment (`_ = foo`)
 	struct ast_expression *object, *value;
 };
 
@@ -175,14 +174,15 @@ struct ast_expression_binarithm {
 	struct ast_expression *lvalue, *rvalue;
 };
 
-struct ast_binding_unpack {
-	struct ident *name;
-	struct ast_binding_unpack *next;
+struct ast_binding_names {
+	struct ident *name; // NULL for _
+	struct ast_binding_names *next;
 };
 
 struct ast_expression_binding {
-	struct ident *name;
-	struct ast_binding_unpack *unpack;
+	// more than one name means tuple unpacking,
+	// otherwise it's a regular binding
+	struct ast_binding_names names;
 	struct ast_type *type;
 	bool is_static;
 	struct ast_expression *initializer;
@@ -221,8 +221,8 @@ struct ast_expression_literal {
 };
 
 struct ast_expression_control {
-	const char *label;
-	struct ast_expression *value; // Only set for yield
+	const char *label; // Never set for return.
+	struct ast_expression *value; // Never set for continue
 };
 
 struct ast_expression_defer {
@@ -241,6 +241,7 @@ struct ast_expression_for {
 	struct ast_expression *cond;
 	struct ast_expression *afterthought;
 	struct ast_expression *body;
+	struct ast_expression *else_branch;
 };
 
 struct ast_expression_free {
@@ -292,10 +293,6 @@ struct ast_expression_propagate {
 	bool abort;
 };
 
-struct ast_expression_return {
-	struct ast_expression *value;
-};
-
 struct ast_expression_slice {
 	struct ast_expression *object;
 	struct ast_expression *start, *end;
@@ -327,6 +324,7 @@ struct ast_field_value {
 
 struct ast_expression_struct {
 	bool autofill;
+	bool undefined;
 	struct ident *type;
 	struct ast_field_value *fields;
 };
@@ -370,7 +368,6 @@ struct ast_expression {
 		struct ast_expression_match match;
 		struct ast_expression_measure measure;
 		struct ast_expression_propagate propagate;
-		struct ast_expression_return _return;
 		struct ast_expression_slice slice;
 		struct ast_expression_struct _struct;
 		struct ast_expression_switch _switch;
@@ -386,13 +383,11 @@ struct ast_global_decl {
 	struct ident *ident;
 	struct ast_type *type;
 	struct ast_expression *init;
-	struct ast_global_decl *next;
 };
 
 struct ast_type_decl {
 	struct ident *ident;
 	struct ast_type *type;
-	struct ast_type_decl *next;
 };
 
 enum func_decl_flags {
