@@ -38,6 +38,7 @@ enum expr_type {
 	EXPR_SWITCH,
 	EXPR_TUPLE,
 	EXPR_UNARITHM,
+	EXPR_UNDEFINED,
 	EXPR_VAARG,
 	EXPR_VAEND,
 	EXPR_VASTART,
@@ -245,7 +246,7 @@ struct expression_literal {
 struct expression_control {
 	const char *label;
 	const struct scope *scope;
-	struct expression *value; // Only set for yield
+	struct expression *value; // Only set for yield and break
 };
 
 struct expression_defer {
@@ -273,6 +274,7 @@ struct expression_for {
 	struct expression *cond;
 	struct expression *afterthought;
 	struct expression *body;
+	struct expression *else_branch;
 };
 
 struct expression_free {
@@ -334,7 +336,7 @@ struct expr_struct_field {
 
 struct expression_struct {
 	struct expr_struct_field *fields;
-	bool autofill;
+	bool autofill, undefined;
 };
 
 struct expression_tuple {
@@ -395,5 +397,6 @@ struct expression {
 };
 
 uint32_t expr_hash(const struct expression *expr);
+bool expr_equal(const struct expression *a, const struct expression *b);
 
 #endif
