@@ -11,11 +11,11 @@
 enum lexical_token {
 	T_ATTR_FINI,
 	T_ATTR_INIT,
-	T_ATTR_OFFSET,
 	T_ATTR_PACKED,
 	T_ATTR_SYMBOL,
 	T_ATTR_TEST,
 	T_ATTR_THREADLOCAL,
+	T_ATTR_UNDEFINED,
 	T_UNDERSCORE,
 	T_ABORT,
 	T_ALIGN,
