@@ -1,11 +1,11 @@
 #ifndef HARE_MOD_H
 #define HARE_MOD_H
 
-struct ast_global_decl;
+struct ast_decls;
 struct context;
 struct ident;
 
 struct scope *module_resolve(struct context *ctx,
-	const struct ast_global_decl *defines, struct ident *ident);
+	const struct ast_decls *defines, struct ident *ident);
 
 #endif
