@@ -49,21 +49,18 @@ enum scope_class {
 	SCOPE_DEFINES,
 };
 
-struct yield {
+struct yield { // and break
 	struct expression **expression;
 	struct yield *next;
 };
 
 struct scope {
-	// Used for for loops
-	bool has_break;
-
 	enum scope_class class;
 	const char *label;
 	struct scope *parent;
 
 	const struct type *hint;
-	struct type_tagged_union *results;
+	struct type_tagged_union results;
 	struct yield *yields;
 
 	// Linked list in insertion order
