@@ -102,11 +102,12 @@ itrunc(struct context *ctx, const struct type *type, uint64_t val)
 		return val;
 	case STORAGE_NULL:
 		return (uintptr_t)NULL;
+	case STORAGE_ERROR:
 	case STORAGE_ALIAS:
 		return itrunc(ctx, type_dealias(ctx, type), val);
 	case STORAGE_ENUM:
 		return itrunc(ctx, type->alias.type, val);
-	case STORAGE_ERROR:
+	case STORAGE_INVALID:
 		return val;
 	case STORAGE_BOOL:
 	case STORAGE_DONE:
@@ -126,6 +127,7 @@ itrunc(struct context *ctx, const struct type *type, uint64_t val)
 	case STORAGE_UNION:
 	case STORAGE_VALIST:
 	case STORAGE_VOID:
+	case STORAGE_UNDEFINED:
 		assert(0);
 	}
 	assert(0);
@@ -409,6 +411,7 @@ eval_literal(struct context *ctx,
 	switch (storage) {
 	case STORAGE_ALIAS:
 	case STORAGE_ENUM:
+	case STORAGE_ERROR:
 		assert(0); // Handled above
 	case STORAGE_ARRAY:;
 		struct array_literal **anext = &out->literal.array;
@@ -467,7 +470,7 @@ eval_literal(struct context *ctx,
 		break;
 	case STORAGE_BOOL:
 	case STORAGE_DONE:
-	case STORAGE_ERROR:
+	case STORAGE_INVALID:
 	case STORAGE_F64:
 	case STORAGE_FCONST:
 	case STORAGE_NOMEM:
@@ -501,6 +504,7 @@ eval_literal(struct context *ctx,
 	case STORAGE_NEVER:
 	case STORAGE_OPAQUE:
 	case STORAGE_VALIST:
+	case STORAGE_UNDEFINED:
 		abort(); // Invariant
 	}
 	return true;
@@ -584,7 +588,11 @@ eval_cast(struct context *ctx,
 		return true;
 	}
 
-	if (from->storage == STORAGE_ERROR) {
+	if (from->storage == STORAGE_INVALID) {
+		return true;
+	} else if (from->storage == STORAGE_UNDEFINED) {
+		out->type = EXPR_UNDEFINED;
+		out->result = to;
 		return true;
 	} else if (from->storage == STORAGE_TAGGED) {
 		out->literal = val.literal.tagged.value->literal;
@@ -674,6 +682,7 @@ eval_cast(struct context *ctx,
 		return true;
 	case STORAGE_NULL:
 	case STORAGE_ALIAS:
+	case STORAGE_ERROR:
 		assert(0); // Handled above
 	case STORAGE_BOOL:
 	case STORAGE_FUNCTION:
@@ -686,9 +695,10 @@ eval_cast(struct context *ctx,
 	case STORAGE_VALIST:
 		assert(0); // Invariant
 	case STORAGE_DONE:
-	case STORAGE_ERROR:
+	case STORAGE_INVALID:
 	case STORAGE_NOMEM:
 	case STORAGE_VOID:
+	case STORAGE_UNDEFINED:
 		return true;
 	}
 
@@ -717,7 +727,7 @@ eval_len(struct context *ctx,
 	case STORAGE_STRING:
 		out->literal.uval = obj.literal.string.len;
 		return true;
-	case STORAGE_ERROR:
+	case STORAGE_INVALID:
 		out->literal.uval = 0;
 		return true;
 	case STORAGE_ARRAY:
@@ -740,7 +750,7 @@ literal_default(struct context *ctx, struct expression *v)
 	struct expression b = {0};
 	const struct type *t = type_dealias(ctx, v->result);
 	switch (t->storage) {
-	case STORAGE_ERROR:
+	case STORAGE_INVALID:
 	case STORAGE_POINTER:
 	case STORAGE_I16:
 	case STORAGE_I32:
@@ -764,6 +774,7 @@ literal_default(struct context *ctx, struct expression *v)
 	case STORAGE_RUNE:
 	case STORAGE_SLICE:
 	case STORAGE_BOOL:
+	case STORAGE_UNDEFINED:
 		break; // calloc does this for us
 	case STORAGE_STRUCT:
 	case STORAGE_UNION:
@@ -812,6 +823,7 @@ literal_default(struct context *ctx, struct expression *v)
 		}
 		break;
 	case STORAGE_ALIAS:
+	case STORAGE_ERROR:
 	case STORAGE_FUNCTION:
 	case STORAGE_NEVER:
 	case STORAGE_OPAQUE:
@@ -851,14 +863,19 @@ count_struct_fields(struct context *ctx, const struct type *type)
 }
 
 static bool
-autofill_struct(struct context *ctx, const struct type *type, struct struct_literal **fields)
-{
+autofill_struct(
+	struct context *ctx,
+	const struct type *type,
+	struct struct_literal **fields,
+	bool undefined
+) {
 	assert(type->storage == STORAGE_STRUCT || type->storage == STORAGE_UNION);
 	for (const struct struct_field *field = type->struct_union.fields;
 			field; field = field->next) {
 		if (!field->name) {
 			bool r = autofill_struct(ctx,
-				type_dealias(ctx, field->type), fields);
+				type_dealias(ctx, field->type),
+				fields, undefined);
 			if (!r) {
 				return false;
 			}
@@ -876,12 +893,16 @@ autofill_struct(struct context *ctx, const struct type *type, struct struct_lite
 			fields[i] = xcalloc(1, sizeof(struct struct_literal));
 			fields[i]->field = field;
 			fields[i]->value = xcalloc(1, sizeof(struct expression));
-			fields[i]->value->type = EXPR_LITERAL;
 			fields[i]->value->result = field->type;
-			// TODO: there should probably be a better error message
-			// when this happens
 			if (!literal_default(ctx, fields[i]->value)) {
-				return false;
+				// TODO: there should probably be a better
+				// error message when this happens
+				if (!undefined) {
+					return false;
+				}
+				fields[i]->value->type = EXPR_UNDEFINED;
+			} else {
+				fields[i]->value->type = EXPR_LITERAL;
 			}
 		}
 	}
@@ -919,7 +940,7 @@ eval_struct(struct context *ctx,
 	assert(in->_struct.autofill || i == n);
 
 	if (in->_struct.autofill) {
-		if (!autofill_struct(ctx, type, fields)) {
+		if (!autofill_struct(ctx, type, fields, in->_struct.undefined)) {
 			return false;
 		}
 	}
@@ -1127,8 +1148,7 @@ eval_address_other(struct context *ctx,
 		return false;
 	}
 
-	static int serial = 0;
-	char *symbol = gen_name(&serial, "static.%d");
+	char *symbol = gen_name(&ctx->id, "static.%d");
 	struct ident *name = mkident(ctx, NULL, symbol);
 
 	append_decl(ctx, &(struct declaration){
@@ -1166,9 +1186,9 @@ eval_unarithm(struct context *ctx,
 	struct expression *restrict out)
 {
 	if (in->unarithm.op == UN_ADDRESS) {
-		if (in->unarithm.operand->result == &builtin_type_error) {
+		if (in->unarithm.operand->result == &builtin_type_invalid) {
 			out->type = EXPR_LITERAL;
-			out->result = &builtin_type_error;
+			out->result = &builtin_type_invalid;
 			out->literal.uval = 0;
 			return true;
 		}
@@ -1246,6 +1266,10 @@ eval_expr(struct context *ctx,
 		return eval_tuple(ctx, in, out);
 	case EXPR_UNARITHM:
 		return eval_unarithm(ctx, in, out);
+	case EXPR_UNDEFINED:
+		out->type = EXPR_UNDEFINED;
+		out->result = &builtin_type_undefined;
+		return true;
 	case EXPR_ALLOC:
 	case EXPR_APPEND:
 	case EXPR_ASSERT:
