@@ -18,11 +18,11 @@ static const char *tokens[] = {
 	// Must match enum lexical_token (lex.h)
 	[T_ATTR_FINI] = "@fini",
 	[T_ATTR_INIT] = "@init",
-	[T_ATTR_OFFSET] = "@offset",
 	[T_ATTR_PACKED] = "@packed",
 	[T_ATTR_SYMBOL] = "@symbol",
 	[T_ATTR_TEST] = "@test",
 	[T_ATTR_THREADLOCAL] = "@threadlocal",
+	[T_ATTR_UNDEFINED] = "@undefined",
 	[T_UNDERSCORE] = "_",
 	[T_ABORT] = "abort",
 	[T_ALIGN] = "align",
@@ -1259,6 +1259,7 @@ token_str(const struct token *tok)
 		case STORAGE_DONE:
 		case STORAGE_ENUM:
 		case STORAGE_ERROR:
+		case STORAGE_INVALID:
 		case STORAGE_FUNCTION:
 		case STORAGE_NEVER:
 		case STORAGE_NOMEM:
@@ -1273,6 +1274,7 @@ token_str(const struct token *tok)
 		case STORAGE_UNION:
 		case STORAGE_VALIST:
 		case STORAGE_VOID:
+		case STORAGE_UNDEFINED:
 			assert(0);
 		}
 		return buf;
