@@ -87,6 +87,7 @@ synassert(bool cond, struct token *tok, ...)
 	}
 }
 
+// Error if the next token is not the provided type.
 static void
 want(struct lexer *lexer, enum lexical_token ltok, struct token *tok)
 {
@@ -96,6 +97,18 @@ want(struct lexer *lexer, enum lexical_token ltok, struct token *tok)
 	synassert(out->token == ltok, out, ltok, T_EOF);
 }
 
+// Advance the lexer if the next token is the provided type. Returns whether or
+// not it advanced.
+static bool
+try(struct lexer *lexer, enum lexical_token ltok)
+{
+	struct token tok = {0};
+	if (lex(lexer, &tok) != ltok) {
+		unlex(lexer, &tok);
+	}
+	return tok.token == ltok;
+}
+
 static struct ast_expression *
 mkexpr(struct location loc)
 {
@@ -122,6 +135,7 @@ mkfuncparams(struct location loc)
 }
 
 static struct ast_expression *parse_statement(struct lexer *lexer);
+static struct ast_expression *parse_initializer(struct lexer *lexer);
 
 struct ident *
 parse_identifier(struct lexer *lexer, const char *name, bool *trailing)
@@ -213,15 +227,10 @@ parse_import(struct lexer *lexer, struct ast_imports *import)
 			synerr(&tok, T_LBRACE, T_TIMES, T_EOF);
 		}
 	} else if (import->ident->ns == NULL) {
-		switch (lex(lexer, &tok)) {
-		case T_EQUAL:
+		if (try(lexer, T_EQUAL)) {
 			import->mode = IMPORT_ALIAS;
 			import->alias = import->ident->name;
 			import->ident = parse_identifier(lexer, NULL, NULL);
-			break;
-		default:
-			unlex(lexer, &tok);
-			break;
 		}
 	}
 }
@@ -229,24 +238,19 @@ parse_import(struct lexer *lexer, struct ast_imports *import)
 static void
 parse_imports(struct lexer *lexer, struct ast_subunit *subunit)
 {
-	struct token tok = {0};
 	struct ast_imports **next = &subunit->imports;
 
 	bool more = true;
 	while (more) {
 		struct ast_imports *imports;
-		switch (lex(lexer, &tok)) {
-		case T_USE:
+		if (try(lexer, T_USE)) {
 			imports = xcalloc(1, sizeof(struct ast_imports));
 			parse_import(lexer, imports);
 			want(lexer, T_SEMICOLON, NULL);
 			*next = imports;
 			next = &imports->next;
-			break;
-		default:
-			unlex(lexer, &tok);
+		} else {
 			more = false;
-			break;
 		}
 	}
 }
@@ -254,26 +258,21 @@ parse_imports(struct lexer *lexer, struct ast_subunit *subunit)
 static void
 parse_parameter_list(struct lexer *lexer, struct ast_function_type *type)
 {
-	struct token tok = {0}, tok2 = {0};
+	struct token tok = {0};
 	want(lexer, T_LPAREN, NULL);
 	type->params = mkfuncparams(lexer->loc);
 	struct ast_function_parameters **next = &type->params;
 	for (;;) {
 		switch (lex(lexer, &tok)) {
 		case T_NAME:;
-			switch (lex(lexer, &tok2)) {
-			case T_COLON:
-				(*next)->name = intern_name(lexer->itbl, tok.name);
-				(*next)->type = parse_type(lexer);
-				break;
-			default:
-				unlex(lexer, &tok2);
-				(*next)->type = mktype(tok.loc);
-				(*next)->type->storage = STORAGE_ALIAS;
-				(*next)->type->alias = parse_identifier(lexer,
-					tok.name, NULL);
-				break;
-			}
+			(*next)->name = intern_name(lexer->itbl, tok.name);
+			want(lexer, T_COLON, NULL);
+			(*next)->type = parse_type(lexer);
+			break;
+		case T_UNDERSCORE:
+			(*next)->name = intern_name(lexer->itbl, "_");
+			want(lexer, T_COLON, NULL);
+			(*next)->type = parse_type(lexer);
 			break;
 		case T_ELLIPSIS:
 			free(*next);
@@ -286,18 +285,12 @@ parse_parameter_list(struct lexer *lexer, struct ast_function_type *type)
 			*next = NULL;
 			return;
 		default:
-			unlex(lexer, &tok);
-			(*next)->type = parse_type(lexer);
-			break;
+			synerr(&tok, T_NAME, T_UNDERSCORE, T_ELLIPSIS, T_RPAREN,
+				T_EOF);
 		}
 
-		switch (lex(lexer, &tok)) {
-		case T_EQUAL:
+		if (try(lexer, T_EQUAL)) {
 			(*next)->default_value = parse_expression(lexer);
-			break;
-		default:
-			unlex(lexer, &tok);
-			break;
 		}
 
 		switch (lex(lexer, &tok)) {
@@ -419,7 +412,7 @@ parse_primitive_type(struct lexer *lexer)
 }
 
 static void parse_binding_unpack(struct lexer *lexer,
-	 struct ast_binding_unpack **next);
+	 struct ast_binding_names *next);
 static struct ast_expression *parse_binding_list(
 		struct lexer *lexer, bool is_static);
 static struct ast_expression *parse_object_selector(struct lexer *lexer);
@@ -464,10 +457,8 @@ parse_enum_type(struct ident *ident, struct lexer *lexer)
 		want(lexer, T_NAME, &tok);
 		(*next)->name = intern_name(lexer->itbl, tok.name);
 		(*next)->loc = tok.loc;
-		if (lex(lexer, &tok) == T_EQUAL) {
+		if (try(lexer, T_EQUAL)) {
 			(*next)->value = parse_expression(lexer);
-		} else {
-			unlex(lexer, &tok);
 		}
 		next = &(*next)->next;
 		switch (lex(lexer, &tok)) {
@@ -494,6 +485,7 @@ parse_struct_union_type(struct lexer *lexer)
 	switch (lex(lexer, &tok)) {
 	case T_STRUCT:
 		type->storage = STORAGE_STRUCT;
+		type->struct_union.packed = try(lexer, T_ATTR_PACKED);
 		break;
 	case T_UNION:
 		type->storage = STORAGE_UNION;
@@ -502,22 +494,14 @@ parse_struct_union_type(struct lexer *lexer)
 		synerr(&tok, T_STRUCT, T_UNION, T_EOF);
 		break;
 	}
-	if (lex(lexer, &tok) == T_ATTR_PACKED) {
-		type->struct_union.packed = true;
-	} else {
-		unlex(lexer, &tok);
-	}
 	want(lexer, T_LBRACE, NULL);
 	while (tok.token != T_RBRACE) {
-		if (lex(lexer, &tok) == T_ATTR_OFFSET) {
-			want(lexer, T_LPAREN, NULL);
-			next->offset = parse_expression(lexer);
-			want(lexer, T_RPAREN, NULL);
-		} else {
-			unlex(lexer, &tok);
-		}
-
 		switch (lex(lexer, &tok)) {
+		case T_UNDERSCORE:
+			want(lexer, T_COLON, NULL);
+			next->name = "_";
+			next->type = parse_type(lexer);
+			break;
 		case T_NAME:;
 			const char *name = tok.name;
 			struct location loc = tok.loc;
@@ -530,7 +514,6 @@ parse_struct_union_type(struct lexer *lexer)
 				unlex(lexer, &tok);
 				next->type = mktype(loc);
 				next->type->storage = STORAGE_ALIAS;
-				next->type->unwrap = false;
 				next->type->alias = parse_identifier(lexer, name, NULL);
 				break;
 			}
@@ -542,7 +525,7 @@ parse_struct_union_type(struct lexer *lexer)
 			next->type = parse_type(lexer);
 			break;
 		default:
-			synerr(&tok, T_NAME, T_STRUCT, T_UNION, T_EOF);
+			synerr(&tok, T_NAME, T_UNDERSCORE, T_STRUCT, T_UNION, T_EOF);
 		}
 		switch (lex(lexer, &tok)) {
 		case T_COMMA:
@@ -563,17 +546,26 @@ parse_struct_union_type(struct lexer *lexer)
 }
 
 static struct ast_type *
-parse_tagged_type(struct lexer *lexer, struct ast_type *first)
+parse_tagged_type(struct lexer *lexer, struct ast_type *first, bool unwrap)
 {
 	struct ast_type *type = mktype(first->loc);
 	type->storage = STORAGE_TAGGED;
 	struct ast_tagged_union_type *next = &type->tagged;
 	next->type = first;
+	next->unwrap = unwrap;
 	struct token tok = {0};
 	while (tok.token != T_RPAREN) {
 		next->next = xcalloc(1, sizeof(struct ast_tagged_union_type));
 		next = next->next;
+
+		if (lex(lexer, &tok) == T_ELLIPSIS) {
+			next->unwrap = true;
+		} else {
+			unlex(lexer, &tok);
+		}
+
 		next->type = parse_type(lexer);
+
 		switch (lex(lexer, &tok)) {
 		case T_BOR:
 			if (lex(lexer, &tok) != T_RPAREN) {
@@ -619,11 +611,18 @@ parse_tuple_type(struct lexer *lexer, struct ast_type *first)
 static struct ast_type *
 parse_tagged_or_tuple_type(struct lexer *lexer)
 {
-	struct ast_type *type = parse_type(lexer);
 	struct token tok = {0};
+	if (lex(lexer, &tok) == T_ELLIPSIS) {
+		struct ast_type *type = parse_type(lexer);
+		want(lexer, T_BOR, &tok);
+		return parse_tagged_type(lexer, type, true);
+	}
+	unlex(lexer, &tok);
+
+	struct ast_type *type = parse_type(lexer);
 	switch (lex(lexer, &tok)) {
 	case T_BOR:
-		return parse_tagged_type(lexer, type);
+		return parse_tagged_type(lexer, type, false);
 	case T_COMMA:
 		return parse_tuple_type(lexer, type);
 	default:
@@ -636,20 +635,9 @@ struct ast_type *
 parse_type(struct lexer *lexer)
 {
 	struct token tok = {0};
-	uint32_t flags = 0;
-	if (lex(lexer, &tok) != T_CONST) {
-		unlex(lexer, &tok);
-	}
-	switch (lex(lexer, &tok)) {
-	case T_LNOT:
-		flags |= TYPE_ERROR;
-		break;
-	default:
-		unlex(lexer, &tok);
-		break;
-	}
+	try(lexer, T_CONST);
 	struct ast_type *type = NULL;
-	bool nullable = false, unwrap = false;
+	bool nullable = false;
 	switch (lex(lexer, &tok)) {
 	case T_BOOL:
 	case T_DONE:
@@ -677,6 +665,11 @@ parse_type(struct lexer *lexer)
 		unlex(lexer, &tok);
 		type = parse_primitive_type(lexer);
 		break;
+	case T_LNOT:
+		type = mktype(lexer->loc);
+		type->storage = STORAGE_ERROR;
+		type->error = parse_type(lexer);
+		break;
 	case T_NULLABLE:
 		nullable = true;
 		want(lexer, T_TIMES, NULL);
@@ -729,22 +722,16 @@ parse_type(struct lexer *lexer)
 		type->storage = STORAGE_FUNCTION;
 		parse_prototype(lexer, &type->func);
 		break;
-	case T_ELLIPSIS:
-		unwrap = true;
-		want(lexer, T_NAME, &tok);
-		// Fallthrough
 	case T_NAME:
 		unlex(lexer, &tok);
 		type = mktype(lexer->loc);
 		type->storage = STORAGE_ALIAS;
-		type->unwrap = unwrap;
 		type->alias = parse_identifier(lexer, NULL, NULL);
 		break;
 	default:
 		synassert_msg(false, "expected type", &tok);
 		break;
 	}
-	type->flags |= flags;
 
 	return type;
 }
@@ -856,10 +843,12 @@ parse_literal(struct lexer *lexer)
 	case STORAGE_STRUCT:
 	case STORAGE_TAGGED:
 	case STORAGE_TUPLE:
+	case STORAGE_UNDEFINED:
 	case STORAGE_UNION:
 	case STORAGE_VALIST:
 		assert(0); // Handled in a different nonterminal
 	case STORAGE_ERROR:
+	case STORAGE_INVALID:
 	case STORAGE_NEVER:
 	case STORAGE_OPAQUE:
 		assert(0); // Invariant
@@ -883,7 +872,7 @@ parse_array_literal(struct lexer *lexer)
 		unlex(lexer, &tok);
 
 		item = *next = xcalloc(1, sizeof(struct ast_expression_list));
-		item->expr = parse_expression(lexer);
+		item->expr = parse_initializer(lexer);
 		next = &item->next;
 
 		switch (lex(lexer, &tok)) {
@@ -923,7 +912,7 @@ parse_field_value(struct lexer *lexer)
 			/* fallthrough */
 		case T_EQUAL:
 			exp->name = name;
-			exp->initializer = parse_expression(lexer);
+			exp->initializer = parse_initializer(lexer);
 			break;
 		default:
 			unlex(lexer, &tok);
@@ -952,6 +941,12 @@ parse_struct_literal(struct lexer *lexer, struct ident *ident)
 	struct token tok = {0};
 	while (tok.token != T_RBRACE) {
 		switch (lex(lexer, &tok)) {
+		case T_ATTR_UNDEFINED:
+			if (lex(lexer, &tok) != T_ELLIPSIS) {
+				synassert(ident != NULL, &tok, T_ELLIPSIS, T_EOF);
+			}
+			exp->_struct.undefined = true;
+			/* fallthrough */
 		case T_ELLIPSIS:
 			synassert(ident != NULL, &tok, T_RBRACE, T_EOF);
 			exp->_struct.autofill = true;
@@ -998,17 +993,16 @@ parse_tuple_expression(struct lexer *lexer, struct ast_expression *first)
 	tuple = tuple->next;
 
 	while (more) {
-		tuple->expr = parse_expression(lexer);
+		tuple->expr = parse_initializer(lexer);
 
 		switch (lex(lexer, &tok)) {
 		case T_RPAREN:
 			more = false;
 			break;
 		case T_COMMA:
-			if (lex(lexer, &tok) == T_RPAREN) {
+			if (try(lexer, T_RPAREN)) {
 				more = false;
 			} else {
-				unlex(lexer, &tok);
 				tuple->next = xcalloc(1,
 					sizeof(struct ast_expression_tuple));
 				tuple = tuple->next;
@@ -1093,17 +1087,14 @@ parse_assertion(struct lexer *lexer, bool is_static,
 	case T_ASSERT:
 		want(lexer, T_LPAREN, &tok);
 		exp->cond = parse_expression(lexer);
-		if (lex(lexer, &tok) == T_COMMA) {
+		if (try(lexer, T_COMMA)) {
 			exp->message = parse_expression(lexer);
-		} else {
-			unlex(lexer, &tok);
 		}
 		want(lexer, T_RPAREN, &tok);
 		break;
 	case T_ABORT:
 		want(lexer, T_LPAREN, &tok);
-		if (lex(lexer, &tok) != T_RPAREN) {
-			unlex(lexer, &tok);
+		if (!try(lexer, T_RPAREN)) {
 			exp->message = parse_expression(lexer);
 			want(lexer, T_RPAREN, &tok);
 		}
@@ -1202,14 +1193,10 @@ parse_index_slice_expression(struct lexer *lexer, struct ast_expression *lvalue)
 	want(lexer, T_LBRACKET, &tok);
 
 	bool is_slice = false;
-	switch (lex(lexer, &tok)) {
-	case T_DOUBLE_DOT:
+	if (try(lexer, T_DOUBLE_DOT)) {
 		is_slice = true;
-		break;
-	default:
-		unlex(lexer, &tok);
+	} else {
 		start = parse_expression(lexer);
-		break;
 	}
 
 	switch (lex(lexer, &tok)) {
@@ -1237,14 +1224,9 @@ parse_index_slice_expression(struct lexer *lexer, struct ast_expression *lvalue)
 		unlex(lexer, &tok);
 	}
 
-	switch (lex(lexer, &tok)) {
-	case T_RBRACKET:
-		break;
-	default:
-		unlex(lexer, &tok);
+	if (!try(lexer, T_RBRACKET)) {
 		end = parse_expression(lexer);
 		want(lexer, T_RBRACKET, &tok);
-		break;
 	}
 
 	exp->type = EXPR_SLICE;
@@ -1265,7 +1247,7 @@ parse_allocation_expression(struct lexer *lexer)
 		exp->type = EXPR_ALLOC;
 		exp->alloc.kind = ALLOC_OBJECT;
 		want(lexer, T_LPAREN, NULL);
-		exp->alloc.init = parse_expression(lexer);
+		exp->alloc.init = parse_initializer(lexer);
 		switch (lex(lexer, &tok)) {
 		case T_COMMA:
 			// alloc(init, cap)
@@ -1313,7 +1295,7 @@ parse_append_insert(struct lexer *lexer, struct location loc,
 			"syntax error: expected indexing expression");
 	}
 	want(lexer, T_COMMA, NULL);
-	expr->append.value = parse_expression(lexer);
+	expr->append.value = parse_initializer(lexer);
 	expr->append.is_static = is_static;
 
 	switch (lex(lexer, &tok)) {
@@ -1624,11 +1606,10 @@ parse_cast_expression(struct lexer *lexer, struct ast_expression *value)
 	if (kind == C_CAST) {
 		exp->cast.type = parse_type(lexer);
 	} else {
-		if (lex(lexer, &tok) == T_NULL) {
+		if (try(lexer, T_NULL)) {
 			exp->cast.type = mktype(tok.loc);
 			exp->cast.type->storage = STORAGE_NULL;
 		} else {
-			unlex(lexer, &tok);
 			exp->cast.type = parse_type(lexer);
 		}
 	}
@@ -1770,13 +1751,8 @@ parse_if_expression(struct lexer *lexer)
 
 	exp->_if.true_branch = parse_expression(lexer);
 
-	switch (lex(lexer, &tok)) {
-	case T_ELSE:
+	if (try(lexer, T_ELSE)) {
 		exp->_if.false_branch = parse_expression(lexer);
-		break;
-	default:
-		unlex(lexer, &tok);
-		break;
 	}
 	return exp;
 }
@@ -1806,19 +1782,21 @@ static void parse_for_predicate(struct lexer *lexer,
 		while (true) {
 			switch (lex(lexer, &tok)) {
 			case T_NAME:
-				binding->name = intern_name(lexer->itbl, tok.name);
+				binding->names.name =
+					intern_name(lexer->itbl, tok.name);
+				break;
+			case T_UNDERSCORE:
 				break;
 			case T_LPAREN:
-				parse_binding_unpack(lexer, &binding->unpack);
+				parse_binding_unpack(lexer, &binding->names);
 				break;
 			default:
-				synerr(&tok, T_NAME, T_LPAREN, T_EOF);
+				synerr(&tok, T_NAME, T_UNDERSCORE,
+					T_LPAREN, T_EOF);
 			}
 
-			if (lex(lexer, &tok) == T_COLON) {
+			if (try(lexer, T_COLON)) {
 				binding->type = parse_type(lexer);
-			} else {
-				unlex(lexer, &tok);
 			}
 
 			if (for_kind_found) {
@@ -1850,8 +1828,7 @@ static void parse_for_predicate(struct lexer *lexer,
 			if (for_exp->kind != FOR_ACCUMULATOR) {
 				return;
 			}
-			if (lex(lexer, &tok) != T_COMMA) {
-				unlex(lexer, &tok);
+			if (!try(lexer, T_COMMA)) {
 				break;
 			}
 			binding->next = xcalloc(1, sizeof(struct ast_expression_binding));
@@ -1862,8 +1839,7 @@ static void parse_for_predicate(struct lexer *lexer,
 		for_exp->cond = parse_expression(lexer);
 	}
 
-	if (lex(lexer, &tok) != T_SEMICOLON) {
-		unlex(lexer, &tok);
+	if (!try(lexer, T_SEMICOLON)) {
 		return;
 	}
 
@@ -1895,6 +1871,15 @@ parse_for_expression(struct lexer *lexer)
 	want(lexer, T_RPAREN, &tok);
 
 	exp->_for.body = parse_expression(lexer);
+
+	switch (lex(lexer, &tok)) {
+	case T_ELSE:
+		exp->_for.else_branch = parse_expression(lexer);
+		break;
+	default:
+		unlex(lexer, &tok);
+		break;
+	}
 	return exp;
 }
 
@@ -1902,12 +1887,8 @@ static struct ast_case_option *
 parse_case_options(struct lexer *lexer)
 {
 	struct token tok = {0};
-	switch (lex(lexer, &tok)) {
-	case T_ARROW:
+	if (try(lexer, T_ARROW)) {
 		return NULL; // Default case
-	default:
-		unlex(lexer, &tok);
-		break;
 	}
 
 	bool more = true;
@@ -1918,16 +1899,12 @@ parse_case_options(struct lexer *lexer)
 		opt->value = parse_expression(lexer);
 		switch (lex(lexer, &tok)) {
 		case T_COMMA:
-			switch (lex(lexer, &tok)) {
-			case T_ARROW:
+			if (try(lexer, T_ARROW)) {
 				more = false;
-				break;
-			default:
-				unlex(lexer, &tok);
+			} else {
 				opt = xcalloc(1, sizeof(struct ast_case_option));
 				*next = opt;
 				next = &opt->next;
-				break;
 			}
 			break;
 		case T_ARROW:
@@ -2128,17 +2105,13 @@ parse_match_expression(struct lexer *lexer)
 }
 
 static void
-parse_binding_unpack(struct lexer *lexer, struct ast_binding_unpack **next)
+parse_binding_unpack(struct lexer *lexer, struct ast_binding_names *next)
 {
 	struct token tok;
 	while (true) {
-		struct ast_binding_unpack *new = xcalloc(1, sizeof *new);
-		*next = new;
-		next = &new->next;
-
 		switch (lex(lexer, &tok)) {
 		case T_NAME:
-			new->name = intern_name(lexer->itbl, tok.name);
+			next->name = intern_name(lexer->itbl, tok.name);
 			break;
 		case T_UNDERSCORE:
 			break;
@@ -2148,6 +2121,13 @@ parse_binding_unpack(struct lexer *lexer, struct ast_binding_unpack **next)
 
 		switch (lex(lexer, &tok)) {
 		case T_COMMA:
+			if (lex(lexer, &tok) == T_RPAREN) {
+				return;
+			}
+			unlex(lexer, &tok);
+			next->next = xcalloc(1,
+				sizeof(struct ast_binding_names));
+			next = next->next;
 			break;
 		case T_RPAREN:
 			return;
@@ -2180,14 +2160,22 @@ parse_binding_list(struct lexer *lexer, bool is_static)
 	do {
 		switch (lex(lexer, &tok)) {
 		case T_NAME:
-			binding->name = intern_name(lexer->itbl, tok.name);
+			binding->names.name =
+				intern_name(lexer->itbl, tok.name);
+			break;
+		case T_UNDERSCORE:
+			if (exp->type != EXPR_BINDING) {
+				error(tok.loc, "can't use underscore with def");
+			}
 			break;
 		case T_LPAREN:
-			synassert(exp->type == EXPR_BINDING, &tok, T_NAME, T_EOF);
-			parse_binding_unpack(lexer, &binding->unpack);
+			if (exp->type != EXPR_BINDING) {
+				error(tok.loc, "can't use tuple unpacking to declare constant");
+			}
+			parse_binding_unpack(lexer, &binding->names);
 			break;
 		default:
-			synerr(&tok, T_NAME, T_LPAREN, T_EOF);
+			synerr(&tok, T_NAME, T_UNDERSCORE, T_LPAREN, T_EOF);
 		}
 		binding->is_static = is_static;
 
@@ -2202,7 +2190,7 @@ parse_binding_list(struct lexer *lexer, bool is_static)
 			synerr(&tok, T_COLON, T_EQUAL, T_EOF);
 		}
 
-		binding->initializer = parse_expression(lexer);
+		binding->initializer = parse_initializer(lexer);
 
 		if (lex(lexer, &tok) == T_COMMA) {
 			binding->next = xcalloc(1, sizeof *binding->next);
@@ -2240,56 +2228,45 @@ static struct ast_expression *
 parse_control_expression(struct lexer *lexer)
 {
 	struct ast_expression *exp = mkexpr(lexer->loc);
+	exp->control.label = NULL;
+	exp->control.value = NULL;
 
 	struct token tok;
 	switch (lex(lexer, &tok)) {
 	case T_BREAK:
+		exp->type = EXPR_BREAK;
+		break;
 	case T_CONTINUE:
-		exp->type = tok.token == T_BREAK ? EXPR_BREAK : EXPR_CONTINUE;
-		exp->control.label = NULL;
-		switch (lex(lexer, &tok)) {
-		case T_COLON:
-			want(lexer, T_NAME, &tok);
-			exp->control.label = tok.name;
-			break;
-		default:
-			unlex(lexer, &tok);
-			break;
-		}
+		exp->type = EXPR_CONTINUE;
 		break;
 	case T_RETURN:
 		exp->type = EXPR_RETURN;
-		exp->_return.value = NULL;
-		switch (lex(lexer, &tok)) {
-		case T_COMMA:
-		case T_ELSE:
-		case T_RBRACE:
-		case T_RBRACKET:
-		case T_RPAREN:
-		case T_SEMICOLON:
-			unlex(lexer, &tok);
-			break;
-		default:
-			unlex(lexer, &tok);
-			exp->_return.value = parse_expression(lexer);
-			break;
-		}
 		break;
 	case T_YIELD:
 		exp->type = EXPR_YIELD;
-		exp->control.value = NULL;
-		switch (lex(lexer, &tok)) {
-		case T_COMMA:
-		case T_ELSE:
-		case T_RBRACE:
-		case T_RBRACKET:
-		case T_RPAREN:
-		case T_SEMICOLON:
+		break;
+	default:
+		abort(); // unreachable
+	}
+
+	switch (lex(lexer, &tok)) {
+	case T_COMMA:
+	case T_ELSE:
+	case T_RBRACE:
+	case T_RBRACKET:
+	case T_RPAREN:
+	case T_SEMICOLON:
+		unlex(lexer, &tok);
+		break;
+	case T_COLON:
+		if (exp->type == EXPR_RETURN) {
+			// return does not take a label
 			unlex(lexer, &tok);
 			break;
-		case T_COLON:
-			want(lexer, T_NAME, &tok);
-			exp->control.label = tok.name;
+		}
+		want(lexer, T_NAME, &tok);
+		exp->control.label = tok.name;
+		if (exp->type == EXPR_YIELD || exp->type == EXPR_BREAK) {
 			switch (lex(lexer, &tok)) {
 			case T_COMMA:
 				exp->control.value = parse_expression(lexer);
@@ -2298,16 +2275,16 @@ parse_control_expression(struct lexer *lexer)
 				unlex(lexer, &tok);
 				break;
 			}
-			break;
-		default:
-			unlex(lexer, &tok);
-			exp->control.value = parse_expression(lexer);
-			break;
 		}
 		break;
 	default:
-		synerr(&tok,
-			T_BREAK, T_CONTINUE, T_RETURN, T_YIELD, T_EOF);
+		unlex(lexer, &tok);
+		if (exp->type == EXPR_CONTINUE) {
+			// continue does not take an expression
+			break;
+		}
+		exp->control.value = parse_expression(lexer);
+		break;
 	}
 	return exp;
 }
@@ -2336,12 +2313,8 @@ parse_compound_expression(struct lexer *lexer)
 	}
 
 	struct location loc = tok.loc;
-	switch (lex(lexer, &tok)) {
-	case T_RBRACE:
+	if (try(lexer, T_RBRACE)) {
 		error(loc, "syntax error: cannot have empty block");
-	default:
-		unlex(lexer, &tok);
-		break;
 	}
 
 	while (true) {
@@ -2367,6 +2340,7 @@ struct ast_expression *
 parse_expression(struct lexer *lexer)
 {
 	struct token tok;
+	struct ast_expression *value;
 	switch (lex(lexer, &tok)) {
 	case T_STATIC:
 		return parse_static_expression(lexer, false);
@@ -2380,12 +2354,15 @@ parse_expression(struct lexer *lexer)
 		return parse_for_expression(lexer);
 	case T_IF:
 		return parse_if_expression(lexer);
+	case T_UNDERSCORE:
+		want(lexer, T_EQUAL, NULL);
+		return parse_assignment(lexer, NULL, BIN_LEQUAL);
 	default:
 		break;
 	}
 
 	unlex(lexer, &tok);
-	struct ast_expression *value = parse_unary_expression(lexer);
+	value = parse_unary_expression(lexer);
 	if (value->type != EXPR_ACCESS && value->type != EXPR_SLICE
 			&& (value->type != EXPR_UNARITHM
 				|| value->unarithm.op != UN_DEREF)) {
@@ -2431,6 +2408,24 @@ parse_expression(struct lexer *lexer)
 	}
 }
 
+static struct ast_expression *
+parse_initializer(struct lexer *lexer)
+{
+	struct token tok;
+	struct ast_expression *expr;
+	switch (lex(lexer, &tok)) {
+	case T_ATTR_UNDEFINED:
+		expr = mkexpr(lexer->loc);
+		expr->type = EXPR_UNDEFINED;
+		return expr;
+	default:
+		unlex(lexer, &tok);
+		break;
+	}
+
+	return parse_expression(lexer);
+}
+
 static struct ast_expression *
 parse_statement(struct lexer *lexer)
 {
@@ -2481,95 +2476,35 @@ parse_global_decl(struct lexer *lexer, enum lexical_token mode,
 		struct ast_global_decl *decl)
 {
 	struct token tok = {0};
-	struct ast_global_decl *i = decl;
 	assert(mode == T_LET || mode == T_CONST || mode == T_DEF);
-	bool more = true;
-	while (more) {
-		if (mode == T_LET || mode == T_CONST) {
-			switch (lex(lexer, &tok)) {
-			case T_ATTR_SYMBOL:
-				i->symbol = parse_attr_symbol(lexer);
-				break;
-			default:
-				unlex(lexer, &tok);
-				break;
-			}
-			switch (lex(lexer, &tok)) {
-			case T_ATTR_THREADLOCAL:
-				i->threadlocal = true;
-				break;
-			default:
-				unlex(lexer, &tok);
-				break;
-			}
-		}
-		i->ident = parse_identifier(lexer, NULL, NULL);
-		switch (lex(lexer, &tok)) {
-		case T_COLON:
-			i->type = parse_type(lexer);
-			if (lex(lexer, &tok) != T_EQUAL) {
-				synassert(mode != T_DEF, &tok, T_EQUAL, T_EOF);
-				unlex(lexer, &tok);
-				break;
-			}
-			/* fallthrough */
-		case T_EQUAL:
-			i->init = parse_expression(lexer);
-			break;
-		default:
-			synerr(&tok, T_EQUAL, T_COLON, T_EOF);
-		}
-
-		switch (lex(lexer, &tok)) {
-		case T_COMMA:
-			lex(lexer, &tok);
-			if (tok.token == T_NAME
-					|| tok.token == T_ATTR_SYMBOL) {
-				i->next = xcalloc(1, sizeof(struct ast_global_decl));
-				i = i->next;
-				unlex(lexer, &tok);
-				break;
-			}
-			/* fallthrough */
-		default:
-			more = false;
+	decl->ident = parse_identifier(lexer, NULL, NULL);
+	switch (lex(lexer, &tok)) {
+	case T_COLON:
+		decl->type = parse_type(lexer);
+		if (lex(lexer, &tok) != T_EQUAL) {
+			synassert(mode != T_DEF, &tok, T_EQUAL, T_EOF);
 			unlex(lexer, &tok);
 			break;
 		}
+		/* fallthrough */
+	case T_EQUAL:
+		decl->init = parse_initializer(lexer);
+		break;
+	default:
+		synerr(&tok, T_EQUAL, T_COLON, T_EOF);
 	}
 }
 
 static void
 parse_type_decl(struct lexer *lexer, struct ast_type_decl *decl)
 {
-	struct token tok = {0};
 	struct ast_type_decl *i = decl;
-	bool more = true;
-	while (more) {
-		i->ident = parse_identifier(lexer, NULL, NULL);
-		want(lexer, T_EQUAL, NULL);
-		switch (lex(lexer, &tok)) {
-		case T_ENUM:
-			i->type = parse_enum_type(i->ident, lexer);
-			break;
-		default:
-			unlex(lexer, &tok);
-			i->type = parse_type(lexer);
-		}
-		switch (lex(lexer, &tok)) {
-		case T_COMMA:
-			if (lex(lexer, &tok) == T_NAME) {
-				i->next = xcalloc(1, sizeof(struct ast_type_decl));
-				i = i->next;
-				unlex(lexer, &tok);
-				break;
-			}
-			/* fallthrough */
-		default:
-			more = false;
-			unlex(lexer, &tok);
-			break;
-		}
+	i->ident = parse_identifier(lexer, NULL, NULL);
+	want(lexer, T_EQUAL, NULL);
+	if (try(lexer, T_ENUM)) {
+		i->type = parse_enum_type(i->ident, lexer);
+	} else {
+		i->type = parse_type(lexer);
 	}
 }
 
@@ -2577,28 +2512,6 @@ static void
 parse_fn_decl(struct lexer *lexer, struct ast_function_decl *decl)
 {
 	struct token tok = {0};
-	bool more = true;
-	while (more) {
-		switch (lex(lexer, &tok)) {
-		case T_ATTR_FINI:
-			decl->flags |= FN_FINI;
-			break;
-		case T_ATTR_INIT:
-			decl->flags |= FN_INIT;
-			break;
-		case T_ATTR_SYMBOL:
-			decl->symbol = parse_attr_symbol(lexer);
-			break;
-		case T_ATTR_TEST:
-			decl->flags |= FN_TEST;
-			break;
-		default:
-			more = false;
-			unlex(lexer, &tok);
-			break;
-		}
-	}
-	want(lexer, T_FN, NULL);
 	decl->ident = parse_identifier(lexer, NULL, NULL);
 	parse_prototype(lexer, &decl->prototype);
 
@@ -2620,25 +2533,65 @@ parse_decl(struct lexer *lexer, struct ast_decl *decl)
 {
 	struct token tok = {0};
 	decl->loc = lexer->loc;
+
+	enum func_decl_flags flags = 0;
+	bool threadlocal = false;
+	const char *symbol = NULL;
+
+	switch (lex(lexer, &tok)) {
+	case T_ATTR_FINI:
+		flags |= FN_FINI;
+		break;
+	case T_ATTR_INIT:
+		flags |= FN_INIT;
+		break;
+	case T_ATTR_SYMBOL:
+		symbol = parse_attr_symbol(lexer);
+		threadlocal = try(lexer, T_ATTR_THREADLOCAL);
+		break;
+	case T_ATTR_TEST:
+		flags |= FN_TEST;
+		break;
+	case T_ATTR_THREADLOCAL:
+		threadlocal = true;
+		break;
+	default:
+		unlex(lexer, &tok);
+		break;
+	}
+
 	switch (lex(lexer, &tok)) {
 	case T_CONST:
 	case T_LET:
+		synassert(flags == 0, &tok, T_FN, T_EOF);
 		decl->decl_type = ADECL_GLOBAL;
+		decl->global.threadlocal = threadlocal;
+		decl->global.symbol = symbol;
 		parse_global_decl(lexer, tok.token, &decl->global);
 		break;
 	case T_DEF:
+		synassert(flags == 0, &tok, T_FN, T_EOF);
+		synassert(symbol == NULL, &tok, T_CONST, T_LET, T_FN, T_EOF);
+		synassert(!threadlocal, &tok, T_CONST, T_LET, T_EOF);
 		decl->decl_type = ADECL_CONST;
 		parse_global_decl(lexer, tok.token, &decl->constant);
 		break;
 	case T_TYPE:
+		synassert(flags == 0, &tok, T_FN, T_EOF);
+		synassert(symbol == NULL, &tok, T_CONST, T_LET, T_FN, T_EOF);
+		synassert(!threadlocal, &tok, T_CONST, T_LET, T_EOF);
 		decl->decl_type = ADECL_TYPE;
 		parse_type_decl(lexer, &decl->type);
 		break;
-	default:
-		unlex(lexer, &tok);
+	case T_FN:
+		synassert(!threadlocal, &tok, T_CONST, T_LET, T_EOF);
+		decl->function.flags = flags;
+		decl->function.symbol = symbol;
 		decl->decl_type = ADECL_FUNC;
 		parse_fn_decl(lexer, &decl->function);
 		break;
+	default:
+		synerr(&tok, T_CONST, T_LET, T_DEF, T_TYPE, T_FN, T_EOF);
 	}
 }
 
