@@ -66,9 +66,11 @@ store_for_type(struct gen_context *ctx, const struct type *type)
 	case STORAGE_ENUM:
 	case STORAGE_ALIAS:
 		return store_for_type(ctx, type->alias.type);
+	case STORAGE_ERROR:
+		return store_for_type(ctx, type->error);
 	case STORAGE_ARRAY:
 	case STORAGE_DONE:
-	case STORAGE_ERROR:
+	case STORAGE_INVALID:
 	case STORAGE_FCONST:
 	case STORAGE_FUNCTION:
 	case STORAGE_ICONST:
@@ -81,6 +83,7 @@ store_for_type(struct gen_context *ctx, const struct type *type)
 	case STORAGE_STRUCT:
 	case STORAGE_TAGGED:
 	case STORAGE_TUPLE:
+	case STORAGE_UNDEFINED:
 	case STORAGE_UNION:
 	case STORAGE_VALIST:
 	case STORAGE_VOID:
@@ -137,8 +140,10 @@ load_for_type(struct gen_context *ctx, const struct type *type)
 	case STORAGE_ENUM:
 	case STORAGE_ALIAS:
 		return load_for_type(ctx, type->alias.type);
-	case STORAGE_ARRAY:
 	case STORAGE_ERROR:
+		return load_for_type(ctx, type->error);
+	case STORAGE_ARRAY:
+	case STORAGE_INVALID:
 	case STORAGE_FCONST:
 	case STORAGE_FUNCTION:
 	case STORAGE_ICONST:
@@ -155,6 +160,7 @@ load_for_type(struct gen_context *ctx, const struct type *type)
 	case STORAGE_VALIST:
 	case STORAGE_VOID:
 	case STORAGE_DONE:
+	case STORAGE_UNDEFINED:
 		abort(); // Invariant
 	}
 	abort(); // Unreachable
