@@ -15,13 +15,13 @@ tagged_qtype(struct gen_context *ctx,
 
 	// Identify maximum alignment among members
 	size_t maxalign = 0, minalign = SIZE_MAX;
-	for (const struct type_tagged_union *tu = &type->tagged;
-			tu; tu = tu->next) {
-		if (maxalign < tu->type->align) {
-			maxalign = tu->type->align;
+	for (size_t i = 0; i < type->tagged.len; i++) {
+		const struct type *t = type->tagged.types[i];
+		if (maxalign < t->align) {
+			maxalign = t->align;
 		}
-		if (minalign > tu->type->align) {
-			minalign = tu->type->align;
+		if (minalign > t->align) {
+			minalign = t->align;
 		}
 	}
 
@@ -29,9 +29,9 @@ tagged_qtype(struct gen_context *ctx,
 	struct qbe_field *field = &def->type.fields;
 	for (size_t align = 1; align <= 8; align <<= 1) {
 		size_t nalign = 0;
-		for (const struct type_tagged_union *tu = &type->tagged;
-				tu; tu = tu->next) {
-			if (tu->type->align != align || tu->type->size == 0) {
+		for (size_t i = 0; i < type->tagged.len; i++) {
+			const struct type *t = type->tagged.types[i];
+			if (t->align != align || t->size == 0) {
 				continue;
 			}
 			++nalign;
@@ -62,13 +62,13 @@ tagged_qtype(struct gen_context *ctx,
 
 		size_t nfield = 0;
 		struct qbe_field *bfield = &values->type.fields;
-		for (const struct type_tagged_union *tu = &type->tagged;
-				tu; tu = tu->next) {
-			if (tu->type->align != align || tu->type->size == 0) {
+		for (size_t i = 0; i < type->tagged.len; i++) {
+			const struct type *t = type->tagged.types[i];
+			if (t->align != align || t->size == 0) {
 				continue;
 			}
 
-			bfield->type = qtype_lookup(ctx, tu->type, true);
+			bfield->type = qtype_lookup(ctx, t, true);
 			bfield->count = 1;
 			if (++nfield < nalign) {
 				bfield->next = xcalloc(1, sizeof(struct qbe_field));
@@ -214,13 +214,14 @@ aggregate_lookup(struct gen_context *ctx, const struct type *type)
 		break;
 	case STORAGE_DONE:
 	case STORAGE_ENUM:
-	case STORAGE_ERROR:
+	case STORAGE_INVALID:
 	case STORAGE_ALIAS:
 	case STORAGE_I8:
 	case STORAGE_U8:
 	case STORAGE_I16:
 	case STORAGE_U16:
 	case STORAGE_BOOL:
+	case STORAGE_ERROR:
 	case STORAGE_I32:
 	case STORAGE_U32:
 	case STORAGE_RCONST:
@@ -243,6 +244,7 @@ aggregate_lookup(struct gen_context *ctx, const struct type *type)
 	case STORAGE_OPAQUE:
 	case STORAGE_NEVER:
 	case STORAGE_NOMEM:
+	case STORAGE_UNDEFINED:
 		abort(); // Invariant
 	}
 
@@ -286,6 +288,8 @@ qtype_lookup(struct gen_context *ctx,
 	case STORAGE_ENUM:
 	case STORAGE_ALIAS:
 		return qtype_lookup(ctx, type->alias.type, xtype);
+	case STORAGE_ERROR:
+		return qtype_lookup(ctx, type->error, xtype);
 	case STORAGE_ARRAY:
 	case STORAGE_SLICE:
 	case STORAGE_STRING:
@@ -298,12 +302,13 @@ qtype_lookup(struct gen_context *ctx,
 		return ctx->arch.ptr;
 	case STORAGE_VALIST:
 		return ctx->arch.ptr;
-	case STORAGE_ERROR:
+	case STORAGE_INVALID:
 	case STORAGE_NEVER:
 	case STORAGE_NOMEM:
 	case STORAGE_OPAQUE:
 	case STORAGE_VOID:
 	case STORAGE_DONE:
+	case STORAGE_UNDEFINED:
 		abort(); // Invariant
 	case STORAGE_FCONST:
 	case STORAGE_ICONST:
@@ -339,12 +344,15 @@ type_is_aggregate(const struct type *type)
 	case STORAGE_UINT:
 	case STORAGE_UINTPTR:
 	case STORAGE_VOID:
+	case STORAGE_UNDEFINED:
 		return false;
 	case STORAGE_FUNCTION:
 		// Special case
 		return false;
 	case STORAGE_ALIAS:
 		return type_is_aggregate(type->alias.type);
+	case STORAGE_ERROR:
+		return type_is_aggregate(type->error);
 	case STORAGE_ARRAY:
 	case STORAGE_SLICE:
 	case STORAGE_STRING:
@@ -359,7 +367,7 @@ type_is_aggregate(const struct type *type)
 	case STORAGE_RCONST:
 		lower_flexible(NULL, type, NULL);
 		return false;
-	case STORAGE_ERROR:
+	case STORAGE_INVALID:
 	case STORAGE_NEVER:
 	case STORAGE_OPAQUE:
 		assert(0); // Invariant
