@@ -194,14 +194,16 @@ emit_literal(const struct expression *expr, FILE *out)
 	case STORAGE_UNION:
 		assert(0); // TODO, blocked on union support in eval
 	case STORAGE_ALIAS:
-	case STORAGE_ERROR:
+	case STORAGE_INVALID:
 	case STORAGE_FUNCTION:
 	case STORAGE_NEVER:
 	case STORAGE_NOMEM:
 		xfprintf(out, "nomem");
 		break;
+	case STORAGE_ERROR:
 	case STORAGE_OPAQUE:
 	case STORAGE_VALIST:
+	case STORAGE_UNDEFINED:
 		assert(0); // Invariant
 	}
 }
@@ -214,9 +216,6 @@ emit_struct(const struct type *type, FILE *out)
 			type->struct_union.packed ? "@packed " : "");
 	for (const struct struct_field *f = type->struct_union.fields;
 			f; f = f->next) {
-		if (!type->struct_union.c_compat) {
-			xfprintf(out, "@offset(%zu) ", f->offset);
-		}
 		if (f->name) {
 			xfprintf(out, "%s: ", f->name);
 		}
@@ -229,15 +228,11 @@ emit_struct(const struct type *type, FILE *out)
 void
 emit_type(const struct type *type, FILE *out)
 {
-	if (type->flags & TYPE_ERROR) {
-		xfprintf(out, "!");
-	}
-
 	char *ident;
 	switch (type->storage) {
 	case STORAGE_BOOL:
 	case STORAGE_DONE:
-	case STORAGE_ERROR:
+	case STORAGE_INVALID:
 	case STORAGE_F32:
 	case STORAGE_F64:
 	case STORAGE_FCONST:
@@ -262,8 +257,13 @@ emit_type(const struct type *type, FILE *out)
 	case STORAGE_UINTPTR:
 	case STORAGE_VALIST:
 	case STORAGE_VOID:
+	case STORAGE_UNDEFINED:
 		xfprintf(out, "%s", type_storage_unparse(type->storage));
 		break;
+	case STORAGE_ERROR:
+		xfprintf(out, "!");
+		emit_type(type->error, out);
+		break;
 	case STORAGE_POINTER:
 		xfprintf(out, "%s*", type->pointer.nullable ? "nullable " : "");
 		emit_type(type->pointer.referent, out);
@@ -287,12 +287,10 @@ emit_type(const struct type *type, FILE *out)
 		break;
 	case STORAGE_TAGGED:
 		xfprintf(out, "(");
-		for (const struct type_tagged_union *tu = &type->tagged;
-				tu; tu = tu->next) {
-			emit_type(tu->type, out);
-			if (tu->next) {
-				xfprintf(out, " | ");
-			}
+		emit_type(type->tagged.types[0], out);
+		for (size_t i = 1; i < type->tagged.len; i++) {
+			xfprintf(out, " | ");
+			emit_type(type->tagged.types[i], out);
 		}
 		xfprintf(out, ")");
 		break;
@@ -304,6 +302,7 @@ emit_type(const struct type *type, FILE *out)
 		xfprintf(out, "fn(");
 		for (const struct type_func_param *param = type->func.params;
 				param; param = param->next) {
+			xfprintf(out, "_: ");
 			if (param->next) {
 				emit_type(param->type, out);
 				xfprintf(out, ", ");
@@ -372,6 +371,7 @@ emit_decl_func(const struct declaration *decl, FILE *out)
 
 	for (const struct type_func_param *param = fntype->func.params;
 			param; param = param->next) {
+		xfprintf(out, "_: ");
 		if (param->next) {
 			emit_type(param->type, out);
 			if (param->default_value) {
@@ -404,13 +404,14 @@ static void
 emit_decl_global(const struct declaration *decl, FILE *out)
 {
 	char *ident = ident_unparse(decl->ident);
-	xfprintf(out, "export let ");
+	xfprintf(out, "export ");
 	if (decl->symbol) {
 		xfprintf(out, "@symbol(\"%s\") ", decl->symbol);
 	}
 	if (decl->global.threadlocal) {
 		xfprintf(out, "@threadlocal ");
 	}
+	xfprintf(out, "let ");
 	xfprintf(out, "%s: ", ident);
 	emit_type(decl->global.type, out);
 	xfprintf(out, ";\n");
